/*
 ============================================================================
 Project     : TinyOS
 Filename    : os.c
 Author      : John Hatch
 Version     :
 Copyright   : Copyright 2013
 Description : OS implementation file for TinyOS
 ============================================================================
 */

#include "lpc2478.h"
#include "print.h"
#include "os.h"

// currentSP is a global variable that holds the value of the
// stack pointer for the current stack.
int currentSP;
TaskBlock* currentTask;

TaskBlock* runList;
TaskBlock* nextTask;

TaskBlock* taskBlockFreeList;
EventBlock* eventBlockFreeList;

// Allocate all TaskBlocks. One for each task.
TaskBlock taskBlocks[NUMBER_OF_TASKS];

// Allocate all the EventBlocks
EventBlock eventBlocks[NUMBER_OF_EVENTS];

// Idle Task pointer and stack
TaskBlock* IDLE_TASK;
int stackIdle[STACKSIZE];
void OSIdleTask(void);

void dumpBlocks(void);
void dumpRunList(void);

void printStringNum(char* str, Int32U num)
{
	printString(str);
	printUint32(num);
	printString("\n");
}

// OSError is where we trap OS errors
void OSError(char* message);

// OS Started is a flag to indicated if OS is ready
Boolean OSStarted;

//
//  Initialize stack.
//  
//  The stack grows down in memory so we want to start by calculating
//  the top of the stack.
//  
//  Then fill in the initial data starting at the top and working down.
//  
//  The important values are the task_address and the initial CPSR.
//  
//  LR doesn't matter since the task is a infinite loop and should not exit.
//  If it doesn't exit, then it doesn't use the LR since it doesn't return.
//  
int InitializeStack(int* stack, void* task_address)
{
	// Calculate the top of the stack.
	int *stkptr = (int*)(stack + STACKSIZE - 1);

	*(stkptr)   = (int)task_address;   // PC
	*(--stkptr) = (int)0x0e0e0e0e;     // R14 - LR
	*(--stkptr) = (int)0x0c0c0c0c;     // R12
	*(--stkptr) = (int)0x0b0b0b0b;     // R11
	*(--stkptr) = (int)0x0a0a0a0a;     // R10
	*(--stkptr) = (int)0x09090909;     // R9
	*(--stkptr) = (int)0x08080808;     // R8
	*(--stkptr) = (int)0x07070707;     // R7
	*(--stkptr) = (int)0x06060606;     // R6
	*(--stkptr) = (int)0x05050505;     // R5
	*(--stkptr) = (int)0x04040404;     // R4
	*(--stkptr) = (int)0x03030303;     // R3
	*(--stkptr) = (int)0x02020202;     // R2
	*(--stkptr) = (int)0x01010101;     // R1
	*(--stkptr) = (int)0x00000000;     // R0
	*(--stkptr) = ARM_SVC_MODE_ARM;    // CPSR

	return (int)(stkptr);
}

//
//  CreateTask initializes a task's stack and its task block.
//
TaskBlock* OSCreateTask(int id, int* stack, void* task_address)
{
	if (taskBlockFreeList != 0)
	{
		// Take a task block from the free list and initialize it
		TaskBlock* newTask = taskBlockFreeList;
		taskBlockFreeList = newTask->next;

		// Initialize the task data
		newTask->stackPointer = InitializeStack(stack, task_address);
		newTask->delayCount = 0;
		newTask->id = id;

		// Push the next task to the front of the run list
		newTask->next = runList;
		newTask->previous = 0;

		if (runList != NULL)
		{
			runList->previous = newTask;
		}

		runList = newTask;
		return newTask;
	}
	else
	{
		OSError("CreateTask can't create new task. No more free task blocks.\n");
		return 0;
	}
}

//
// Initialize OS state variables
//
void OSInit(void)
{
	OSStarted = FALSE;
	runList = NULL;
	nextTask = 0;

	// Initialize blocks with some values to
	// help debugging OS.
	for (int i = 0; i < NUMBER_OF_TASKS; i++)
	{
		taskBlocks[i].debugId = 1000 + i;
		taskBlocks[i].stackPointer = 0xFFFFFFFF;
		taskBlocks[i].id = 9999;
		taskBlocks[i].delayCount = 9999;
		taskBlocks[i].next = (void*)0xFFFFFFFF;
		taskBlocks[i].previous = (void*)0xFFFFFFFF;
	}

	//
	// Add all the TaskBlocks to the task block free list
	//
	taskBlockFreeList = &taskBlocks[0];
	struct _TaskBlock* currentTB = taskBlockFreeList;
	currentTB->previous = NULL;

	for(int i = 1; i < NUMBER_OF_TASKS; i++)
	{
		currentTB->next = &taskBlocks[i];
		taskBlocks[i].previous = currentTB;
		currentTB = &taskBlocks[i];
	}

	currentTB->next = NULL;

	//
	// Add all the EventBlocks to the event block free list
	//
	eventBlockFreeList = &eventBlocks[0];
	EventBlock* currentEB = eventBlockFreeList;

	for(int i = 1; i < NUMBER_OF_TASKS; i++)
	{
		currentEB->waitListHead = 0;
		currentEB->waitListTail = 0;
		currentEB->next = &eventBlocks[i];
		currentEB = &eventBlocks[i];
	}

	currentEB->next = NULL;

	// Create the idle task
	IDLE_TASK = OSCreateTask(0, stackIdle, OSIdleTask);
}

//
//  StartOS initializes the globals used by the scheduler and switches
//  to the first task.
//
void OSStart(void)
{
    // Set current task to the first task on the run list.
    currentTask = runList;
    nextTask = currentTask->next;

    currentSP = currentTask->stackPointer;

    OSStarted = TRUE;

    // Run the current task
    SwitchToCurrent();         // This call never returns to this function.
}

//
// Simple task scheduler that decides which task to run.
// It only considers tasks on the run list.
// It saves the current stack pointer to the task block
// then picks a new task to run. If if cannot find a
// task to run it runs the idleTask.
//
void _OSScheduler(void)
{
	CPU_SR cpu_sr = 0;

	ENTER_CRITICAL();

	//printString("Scheduler\n");

	// Save the current stack pointer for the current task.
	currentTask->stackPointer = currentSP;

    // Start with the nextTask
    TaskBlock* blockPtr = nextTask;

    // Loop through the run list of task blocks to find the next task to run
    while(1)
    {
        // If the task is not the idle task and is not delayed then
    	// it is runnable. We have found our task and can stop looking.
    	if (blockPtr != IDLE_TASK && blockPtr->delayCount == 0)
    	{
    		break;
    	}

    	// Move to the next task in the list
    	blockPtr = blockPtr->next;

    	// If we are at the end then start over
    	if (blockPtr == 0)
    	{
    		blockPtr = runList;
    	}

        // If we loop around back to where we started and have not found
        // a runnable task then stop looking and run the idle task.
        if (blockPtr == nextTask)
        {
        	blockPtr = IDLE_TASK;
        	break;
        }
    }

    // Set current task
    currentTask = blockPtr;
    currentSP = currentTask->stackPointer;

    // Set next task so we will know late even if the currentTask
    // gets block and moves to a wait list.
    nextTask = currentTask->next;

    // If we are at the end then start over
	if (nextTask == 0)
	{
		nextTask = runList;
	}

    EXIT_CRITICAL();
}

//
//  IrqDispatcher looks up the handler for the interrupt source from the VIC's vector table
//  and calls that handler. The VIC needs to be configured before calling this function.
//  The function InitializeTimerTick initialize the VIC for timer0 interrupts.
//
void IrqDispatcher(void)
{
	FNCT_PTR pfnct;

	// Read the interrupt vector from the VIC
	pfnct = (FNCT_PTR)VICADDRESS;

	// Check for null before call it.
	if (pfnct != (FNCT_PTR)0)
	{
		// Call the handler function for the current interrupt
		(*pfnct)();

		// Acknowledge the VIC interrupt
		VICADDRESS = 1;
	}
    else
    {
    	OSError("IrqDispatcher: Interrupt vector not set.\n");
    }
}

//
// TickHandler handles interrupts caused by the timer0.
// It resets the timer then decrements the counter of any
// delayed task. Then calls to scheduler.
//
void OSTickHandler(void)
{
	// Reset timer
	T0IR = 0xFF;

	// If the OS has not been start then just return
	if (OSStarted != TRUE)
		return;

    TaskBlock* blockPtr = runList;

    while (blockPtr != 0)
    {
    	if (blockPtr->delayCount > 0)
    	{
    		blockPtr->delayCount = blockPtr->delayCount - 1;
    	}

    	blockPtr = blockPtr->next;
    }

	_OSScheduler();
}

//
//  DelayTask sets a task's delayCount causing the scheduler to skip
//  the task for the number of ticks specified.
//  DelayTask also calls ContextSwitch which calls the scheduler and
//  forces an immediate context switch to another task.
//
void OSDelayTask(int numOfTicks)
{
	CPU_SR cpu_sr = 0;

	ENTER_CRITICAL();
    currentTask->delayCount = numOfTicks;
    SwitchContext();
	EXIT_CRITICAL();
}

//
// idleTask runs whenever the OS cannot any other tasks.
// For example if all tasks are delayed.
//
void OSIdleTask(void)
{
	//CPU_SR cpu_sr = 0;

	while(1)
	{
		//ENTER_CRITICAL();
		//printString("Idle Task\n");
		//EXIT_CRITICAL();
		//for(int i = 0; i < 10000; i++);    // delay
	}
}

//
// Removes the current task from the runList and return it.
//
TaskBlock* _OSRemoveCurrentTaskFromRunList(void)
{
	TaskBlock* task = currentTask;

	if (runList == task)
	{
		runList = task->next;
	}

	if (task->previous != 0)
	{
		task->previous->next = task->next;
	}

	if (task->next != 0)
	{
		task->next->previous = task->previous;
	}

	return task;
}

//
// Adds a task to the front of the runList.
//
void _OSAddTaskToRunList(TaskBlock* task)
{
	task->next = runList;
	task->previous = 0;

	// Assertion: runList always has at least one task, the idle task.
	runList->previous = task;
	runList = task;
}

//
// Adds a task to the tail of an event's wait list.
//
void _OSAddTaskToEventWaitList(EventBlock* eventBlock, TaskBlock* task)
{
	task->next = 0;
	task->previous = 0;

	if (eventBlock->waitListHead == 0)
	{
		eventBlock->waitListHead = task;
		eventBlock->waitListTail = task;
	}
	else
	{
		eventBlock->waitListTail->next = task;
		eventBlock->waitListTail = task;
	}
}

//
// Push free eventBlock on to the EventBlock free list
//
void _OSAddEventBlockToFreeList(EventBlock* eventBlock)
{
	eventBlock->next = eventBlockFreeList;
	eventBlockFreeList = eventBlock;
}

//
// OSCreateSemaphore task an EventBlock from the free list
// and initializes it with the semaphores count. It then
// returns it to the user. The user is responsible for
// keeping track of the EventBlock and deleting it with
// OSDeleteSemaphore when they are finished with it.
//
EventBlock* OSCreateSemaphore(int count)
{
	CPU_SR cpu_sr = 0;

	ENTER_CRITICAL();
	if (eventBlockFreeList != 0)
	{
		// Take a event block from the free list and initialize it
		EventBlock* newEvent = eventBlockFreeList;
		eventBlockFreeList = newEvent->next;

		newEvent->count = count;
		newEvent->next = 0;
		newEvent->waitListHead = 0;
		newEvent->waitListTail = 0;
		EXIT_CRITICAL();
		return newEvent;
	}
	else
	{
		OSError("OSCreateSemaphore: Can't create new semaphore. No more free Event blocks.\n");
		EXIT_CRITICAL();
		return 0;
	}
}

//
// Pends on a semaphore by trying to acquire a semaphore for the calling task.
// If there are none available it blocks the task by removing it from the runLit
// and adding it to the event's wait list. For now the wait list is First In First Out.
//
void OSPendSemaphore(EventBlock* eventBlock)
{
	CPU_SR cpu_sr = 0;

	ENTER_CRITICAL();
	if (eventBlock->count > 0)
	{
		// Semaphore is count is great then 0 is the task is
		// able to acquire it so decrement the count and return.
		eventBlock->count--;
		EXIT_CRITICAL();
		return;
	}

	// Semaphore is 0 so block the current task by removing it from runlist,
	// adding it to event block wait list then call scheduler.
	TaskBlock* task = _OSRemoveCurrentTaskFromRunList();
	_OSAddTaskToEventWaitList(eventBlock, task);

	// The current task was block so time to call the scheduler
	// and switch context.
	SwitchContext();
	EXIT_CRITICAL();
}

//
// Posts to a semaphore. If there are waiting tasks, the first task will
// acquire the semaphore and be moved to the runList. If there are not
// waiting tasks the semaphore is simply increment.
//
void OSPostSemaphore(EventBlock* eventBlock)
{
	CPU_SR cpu_sr = 0;

	ENTER_CRITICAL();
	//
	// If there are no tasks waiting then just increment the
	// count and return.
	//
	if (eventBlock->waitListHead == 0)
	{
		eventBlock->count++;
		EXIT_CRITICAL();
		return;
	}

	//
	// If there was at least one task, pop the task at the head of
	// the list, add it to the runList and call the scheduler.
	// We don't change the count because the semaphore was posted
	// and acquired at the same time.
	//
	TaskBlock* task = eventBlock->waitListHead;
	eventBlock->waitListHead = task->next;

	if (eventBlock->waitListHead == 0)
	{
		eventBlock->waitListTail = 0;
	}

	_OSAddTaskToRunList(task);

	// A task was unblocked and added to the runlist.
	// Depending on the scheduler policy we might want to call
	// for a reschedule and context switch.
	// SwitchContext();

	EXIT_CRITICAL();
}

//
// Deletes a semaphore by making all waiting tasks runnable then
// returning the EventBlock to the EventBlock free list.
//
void OSDeleteSemaphore(EventBlock* eventBlock)
{
	CPU_SR cpu_sr = 0;
	ENTER_CRITICAL();

	//
	// Make all the waiting tasks runnable by moving them to the runList.
	// Best practice is not to have any waiting task when deleting a
	// semaphore.
	//
	TaskBlock* current = eventBlock->waitListHead;

	while (current != 0)
	{
		TaskBlock* next = current->next;
		_OSAddTaskToRunList(current);
		current = next;
	}

	_OSAddEventBlockToFreeList(eventBlock);

	SwitchContext();
	EXIT_CRITICAL();
}

//
// Print error message and halt the OS
//
void OSError(char* message)
{
	printString(message);

	while(1) {} // Stop here so we can debug the problem instead of crashing.
}

//
// For debugging OS
//
void dumpBlocks(void)
{
	printString("\n\n-----------------------------------");

	for (Int32U i = 0; i < NUMBER_OF_TASKS; i++)
	{
		printString("\n\nBlock: ");
		printUint32(i);
		printString("\n   Address:      ");
		printHex((Int32U)&taskBlocks[i]);
		printString("\n   stackPointer: ");
		printHex(taskBlocks[i].stackPointer);
		printString("\n   id:           ");
		printUint32((Int32U)taskBlocks[i].id);
		printString("\n   debugId:      ");
		printUint32((Int32U)taskBlocks[i].debugId);
		printString("\n   delayCount:   ");
		printUint32((Int32U)taskBlocks[i].delayCount);
		printString("\n   next:         ");
		printHex((Int32U)taskBlocks[i].next);
		printString("\n   previous:     ");
		printHex((Int32U)taskBlocks[i].previous);
	}
	printString("\n");
}

//
// For debugging OS
//
void dumpRunList(void)
{
	printString("\n\n----- RunList ------------------------------");

	TaskBlock* blockPtr = runList;

	while (blockPtr != 0)
	{
		printString("\n---debugId:      ");
		printUint32((Int32U)blockPtr->debugId);
		printString("\n   Address:      ");
		printHex((Int32U)blockPtr);
		printString("\n   stackPointer: ");
		printHex(blockPtr->stackPointer);
		printString("\n   id:           ");
		printUint32((Int32U)blockPtr->id);
		printString("\n   delayCount:   ");
		printUint32((Int32U)blockPtr->delayCount);
		printString("\n   next:         ");
		printHex((Int32U)blockPtr->next);
		printString("\n   previous:     ");
		printHex((Int32U)blockPtr->previous);
		printString("\n");

		blockPtr = blockPtr->next;
	}
	printString("\n");
}

