/*
 ============================================================================
 Name        : shell.c
 Author      : John Hatch
 Version     :
 Copyright   : Copyright 2013
 Description : Tiny shell parser
 ============================================================================
 */

#include "lpc2478.h"
#include "libc.h"
#include "os.h"
#include "shell.h"
#include "print.h"
#include "uart.h"

EventBlock* uartSem;
char commandBuffer[BUFFERSIZE];
unsigned int commandIndex;

typedef int (*CmdFuncPtr)(int, char **);

//
// Command prototypes
//
int Echo(int argc, char **argv);
int Version(int argc, char **argv);
int Help(int argc, char **argv);
int StartCount(int argc, char **argv);
int StopCount(int argc, char **argv);
int ShowCount(int argc, char **argv);

//
// Command table has one entry per command.
// An entry is the commands name and its function pointer.
//
struct _CommandEntry
{
	char* name;
	CmdFuncPtr handler;
} commandTable[] = {
	{"help",       Help},
	{"echo",       Echo},
	{"version",    Version},
	{"startcount", StartCount},
	{"stopcount",  StopCount},
	{"showcount",  ShowCount},
	{NULL,         NULL}
};

///////////////////////////////////////////////////////////////////////////////////////////
//
// ShowArgs prints all the arguments in the command line.
//
int Echo(int argc, char **argv)
{
	printF("argc: %d\n", argc);

	for(int i = 0; i < argc; i++)
	{
		printF("   %s\n", argv[i]);
	}

	return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////
//
// Version prints build date and time.
//
int Version(int argc, char **argv)
{
	char *date = __DATE__;
	char *time = __TIME__;

	printF("TinyOS built on: %s at %s\n", date, time);
	return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////
//
// Help prints the commands in the command table.
//
int Help(int argc, char **argv)
{
	printF("Known commands:\n");

	int i = 0;
	while(commandTable[i].name)
	{
		printF("   %s\n", commandTable[i++].name);
	}
	return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////
//
// LookupCommand finds the command in the command table.
//
CmdFuncPtr LookupCommand(char* commandLine)
{
    char commandName[64];
    struct _CommandEntry *entry;

    strncpy(commandName, commandLine, sizeof commandName);

    commandName[sizeof commandName - 1] = '\0';

    if (strtok(commandName, " \t") == NULL)
    {
        return (NULL);
    }

    for (entry = &commandTable[0]; entry->name != NULL; ++entry)
    {
        if (strcmp(commandName, entry->name) == 0)
        {
            return (entry->handler);
        }
    }
    return (NULL);
}

///////////////////////////////////////////////////////////////////////////////////////////
//
// ExecuteCommandLine parse the command line and executes the command.
//
void ExecuteCommandLine(void)
{
    CmdFuncPtr commandPtr;

    commandIndex = 0;

    while (commandBuffer[commandIndex] == ' ' || commandBuffer[commandIndex] == '\t')
    {
        ++commandIndex;
    }

    if (commandBuffer[commandIndex] == '\0')
    {
        commandIndex = 0;
        commandBuffer[commandIndex] = '\0';
        printF("\n>");
        return;
    }

    commandPtr = LookupCommand(&commandBuffer[commandIndex]);

    if (commandPtr != NULL)
    {
        char *argPtrs[24];
        int   argCount;

        argCount = 0;
        argPtrs[argCount++] = strtok(&commandBuffer[commandIndex], " \t");

        for (; argCount < (int)(sizeof argPtrs / sizeof *argPtrs); ++argCount)
        {
            if ((argPtrs[argCount] = strtok(NULL, " \t")) == NULL)
                break;
        }

        if (commandPtr(argCount, argPtrs) != 0)
        {
            printF("Error occurred.\n");
        }
    }
    else
    {
        printF("Unknown command: '%s'\n", commandBuffer);
    }

    printF(">");

    commandIndex = 0;
    commandBuffer[commandIndex] = '\0';
}

///////////////////////////////////////////////////////////////////////////////////////////
//
// UARTHandler handles in coming character from the UART.
// It stores them in the command buffer until the user
// hits return. On return it signals the line is ready.
//
void UARTHandler(void)
{
	char c = UARTReceive(UART_CH_0);

	if (c == 127)
	{
		// Backspace
		printString("\b \b");
		commandIndex--;
	}
	else if (c == '\r')
	{
		// User has hit return
		printChar('\n');
		commandBuffer[commandIndex] = 0;

		// Signal ShellTask to process the command line.
		OSPostSemaphore(uartSem);
	}
	else
	{
		// Echo character and store it in the buffer.
		printChar(c);
		commandBuffer[commandIndex++] = c;

		if (commandIndex > BUFFERSIZE - 1)
		{
			commandIndex = 0;
		}
	}
}

///////////////////////////////////////////////////////////////////////////////////////////
//
// ShellTask executes the command line when signaled
// by the UARTHandler that the line is ready.
//
int ShellTask(void)
{
	printF("Welcome to Tiny Shell.\n");

	uartSem = OSCreateSemaphore(0);
	UARTInit(38400, UART_CH_0, 0);
	UARTInitInterrupt(UARTHandler);

	printF("Type help to see available commands.\n");
	printF(">");

	while(1)
	{
		OSPendSemaphore(uartSem);
		ExecuteCommandLine();
	};

	printF("Exiting.\n");
	return 0;
}
