
.set INT_DISABLED,           0xC0    /* Disable both FIQ and IRQ. */
.set MODE_IRQ,               0x12    /* IRQ Mode                  */
.set MODE_SVC,               0x13    /* Supervisor Mode           */
.set ARM_THUMB_MODE,         0x20    /* Set THUMB mode. */
.set ARM_ARM_MODE,           0x00    /* Set ARM mode. */


.extern currentSP
.extern IrqDispatcher
.extern _OSScheduler

.global SimpleIRQHandler

.global SwitchToCurrent
.global OSIrqHandler
.global SwitchContext

.code 32

/*
 *  Switch to current task without saving previous context.
 *  Used to run the first task the first time.
 *
 *  Pops the initial context setup up by the initialize_stack function.
 *  That means this function will act like a return to the task.
 *
 */
SwitchToCurrent:
                                      /* Switch to the current task's stack. */
    ldr   r0, =currentSP              /* Load r0 with the address for currentSP */
    ldr   sp, [r0]                    /* Load sp with the value stored at the address */
                                      /* sp now points to the task's stack. */

                                      /* Set SPSR to the CPSR for the task. */
    ldmfd sp!, {r0}                   /* Pop the first value from the stack into r0. */
                                      /* That value was the CPSR for the task. */
    msr   spsr_cxsf, r0               /* Load SPSR with that CPSR value in r0. */

                                      /* Run task. */
    ldmfd sp!, {r0-r12, lr, pc}^      /* Pop the rest of the stack setting regs and pc for the task */
                                      /* Acts like a call the task. */

/*
 *  OS interrupt handler that calls the IRQ dispatcher.
 *
 */
OSIrqHandler:

    /* IRQ mode with IRQ stack */

    sub    lr, lr, #4                      /* LR offset to return from this exception: -4. */
    stmfd  sp!, {r0-r2}                    /* Push working registers r0-r3.  */

    mrs    r0, spsr                        /* Save task's cpsr (stored as the IRQ's spsr) to r0 */
    mov    r1, lr                          /* Save lr which is the task's pc to r1 */

                                           /* We'll use r2 to retrieve the data off of the IRQ stack. */
    mov    r2, sp                          /* Save exception's stack pointer.*/
    add    sp, sp, #(3 * 4)                /* Reset IRQ's stack back to where it was so it will be ready next time. */

    msr     cpsr_c, #(INT_DISABLED | MODE_SVC)  /* Change to SVC mode with interrupts disabled. */

    /* SVC mode with SVC stack. This is the stack of interrupted task. */

    stmfd  sp!, {r1}                       /* Push task's PC */
    stmfd  sp!, {lr}                       /* Push task's LR */
    stmfd  sp!, {r3-r12}                   /* Push task's R4-R12 */

                                           /* Move task's r0-r3 from exception stack to task's stack. */
    ldmfd  r2!, {r3-r5}                    /* By first loading r3-r5 with r0-r2 that where saved on the irq stack */
    stmfd  sp!, {r3-r5}                    /* then by pushing them onto the task's stack */

    stmfd  sp!, {r0}                       /* Push task's CPSR, which was the IRQ mode's SPSR to the task stack. */

    ldr    r0, =currentSP                  /* Save SP to currentSP before calling IrqHandler */
    str    sp, [r0]

    bl     IrqDispatcher                   /* Call the task scheduler which sets currentSP and resets the timer. */

    ldr    r0, =currentSP                  /* Set sp to the next task's sp by reading the value stored in currentSP. */
    ldr    sp, [r0]

    ldmfd  sp!, {r0}                       /* Pop task's CPSR and restore it to spsr. */
    msr    spsr_cxsf, r0                   /* spsr will be moved to cpsr when we pop the context with ldmfd */

    ldmfd  sp!, {r0-r12, lr, pc}^          /* Pop task's context. Restores regs and cpsr which reenables interrupts. */


/*
 * OS context switcher which is called from C code to force a context switch.
 * It saves the current task's context, calls the scheduler
 * then restores the context of the next task.
 *
 */
SwitchContext:

    /* SVC mode with SVC stack */

    stmfd  sp!, {lr}
    stmfd  sp!, {lr}
    stmfd  sp!, {r0-r12}

    mrs    r0, cpsr
    tst    lr, #1                          /* Test if called from Thumb mode */
    orrne  r0, r0, #ARM_THUMB_MODE         /* If yes, then set the T-bit     */
    stmfd  sp!, {r0}

    ldr    r0, =currentSP                  /* Save SP to currentSP before calling IrqHandler */
    str    sp, [r0]

    bl     _OSScheduler                    /* Call the task scheduler which sets currentSP. */

    ldr    r0, =currentSP                  /* Set sp to the new task's sp by reading the value stored in currentSP. */
    ldr    sp, [r0]

    ldmfd  sp!, {r0}                       /* Pop task's CPSR and restore it to spsr. */
    msr    spsr, r0                        /* spsr will be moved to cpsr when we pop the context with ldmfd */

    ldmfd  sp!, {r0-r12, lr, pc}^          /* Pop task's context. Restores regs and cpsr which reenables interrupts. */




